/* ------------------------------------------------------------------------------------------------------------
• FoscTremolo (abjad 3.0)

Tremolo (of exactly two notes).


• Example 1

With three beams by default.

a = FoscChord([61, 64], 1/4);
m = FoscTremolo();
a.attach(m);
a.show;


• Example 2

With two beams.

a = FoscChord([61, 64], 1/4);
m = FoscTremolo(beamCount: 2);
a.attach(m);
a.show;


• Example 3

With slur.

a = FoscChord([61, 64], 1/4);
m = FoscTremolo(isSlurred: true);
a.attach(m);
a.show;
------------------------------------------------------------------------------------------------------------ */
FoscTremolo : Fosc {
    ///////////////////////////////////////////////////////////////////////////////////////////////////////////
    // INIT
    ///////////////////////////////////////////////////////////////////////////////////////////////////////////
    var <beamCount, <isSlurred;
    var <formatSlot;
    *new { |beamCount=3, isSlurred=false|
        assert(beamCount.isInteger);
        assert(beamCount >= 1);
        assert(isSlurred.isKindOf(Boolean));
        ^super.new.init(beamCount, isSlurred);
    }
    init { |argBeamCount, argIsSlurred|
        beamCount = argBeamCount;
        isSlurred = argIsSlurred;
    }
    ///////////////////////////////////////////////////////////////////////////////////////////////////////////
    // PUBLIC INSTANCE PROPERTIES
    ///////////////////////////////////////////////////////////////////////////////////////////////////////////
    /* --------------------------------------------------------------------------------------------------------
    • beamCount

    Gets beam count of tremolo.


    • Example 1

    m = FoscTremolo();
    m.beamCount;


    • Example 2

    m = FoscTremolo(beamCount: 2);
    m.beamCount;
    -------------------------------------------------------------------------------------------------------- */
    /* --------------------------------------------------------------------------------------------------------
    • isSlurred

    Is true when tremolo is slurred.


    • Example 1

    m = FoscTremolo(isSlurred: true);
    m.isSlurred;
    -------------------------------------------------------------------------------------------------------- */
    /* --------------------------------------------------------------------------------------------------------
    • tweaks

    Tweaks are not implemented on tremolo.
        
    The Lilypond '\repeat tremolo' command refuses tweaks.

    Override the Lilypond 'Beam' grob instead.
    -------------------------------------------------------------------------------------------------------- */
    tweaks {
        // pass
    }
    ///////////////////////////////////////////////////////////////////////////////////////////////////////////
    // PUBLIC INSTANCE METHODS: SPECIAL METHODS
    ///////////////////////////////////////////////////////////////////////////////////////////////////////////
    /* --------------------------------------------------------------------------------------------------------
    • format
    
    Formats stem tremolo.
    -------------------------------------------------------------------------------------------------------- */
    /* --------------------------------------------------------------------------------------------------------
    • str

    Gets string representation of tremolo.
    -------------------------------------------------------------------------------------------------------- */
}

